import os
import sys
import xlwings as xw
import pandas as pd
import dreamtools as dt
import re


# workbook_path = "konjunktur.xlsx"
# gdx_path = "../Model/Gdx/static_calibration.gdx"
# sheet_name = "data"

def gdx_to_excel(gdx_path, workbook_path, sheet_name):
    """
    Fill an Excel worksheet with data from a GDX file.

    Parameters
    ----------
    gdx_path : str
        Path to GDX file
    workbook_path : str
        Path to an Excel workbook
    sheet_name : str
        Name of a sheet in the workbook
        In that sheet, the script expects a table with variables as index and years as column headers,
        starting in cell A1.
        The function will then fill the table with data from the GDX file.
        The function assumes that the time index is always last.
    """
    assert os.path.isfile(gdx_path), f"{gdx_path} not found"
    db = dt.Gdx(gdx_path)

    assert os.path.isfile(workbook_path), f"{workbook_path} not found"
    wb = xw.Book(workbook_path)

    assert sheet_name in wb.sheet_names, f'"{sheet_name}" not found in {workbook_path}'
    sheet = wb.sheets[sheet_name]

    data_range = sheet.range("A1", "B2").expand()
    df = data_range.options(pd.DataFrame, index=True).value
    try:
        fill_dataframe(df, db)
        data_range.offset(1, 1).resize(*df.shape).value = df.values
    except Exception as e:
        data_range.offset(1, 1).resize(*df.shape).value = None
        raise e
    
    wb.save()

def get_variable_name(variable_key):
    """Returns the variable name from a variable key, e.g. "qY[tje]" -> "qY" """
    return variable_key.split("[")[0].lower()

def get_index(variable_key):
    """Returns the index from a variable key, e.g. "qY[tje]" -> "tje" """
    if "[" in variable_key:
        index_str = variable_key.split("[")[1][:-1]
        return [x.lower() for x in index_str.split(",")]

def index_to_lowercase(idx):
    """Change a pandas multiindex to lower case only"""
    return idx.set_levels([x.lower() if isinstance(x, str) else x for x in level] for level in idx.levels)

def remove_time_index(variable_key):
    """
    Remove time index from variable key.
    Examples:
    remove_time_index("qY[tje,t]") -> "qY[tje]"
    remove_time_index("qBNP[t]") -> "qBNP"
    """
    variable_key = re.sub(r"\[t\]", "", variable_key)
    variable_key = re.sub(r",t\]", "]", variable_key)
    return variable_key  

def remove_quotes(variable_key):
    """
    Remove quotes from variable key.
    Examples:
    remove_quotes("qY['tje']") -> "qY[tje]"
    """
    return variable_key.replace("'", "").replace('"', "")

def fill_dataframe(df, db):
    """
    Fills a dataframe with data from a GDX database
    The dataframe needs to have variables as index and years as column headers.
    E.g.
                    2016	2050
    vC[cTot]	    833 	1089
    pI_s[iM,tje]	None	None
    pI_s[iB,tje]	None	None
    """
    years = df.columns.astype(int)
    variable_keys = df.index
    for variable_key in variable_keys:
        error_message = None
        try:
            variable_name = get_variable_name(variable_key)
            index = get_index(remove_quotes(remove_time_index(variable_key)))

            if variable_name in [x.lower() for x in db]:
                variable = db[[x for x in db if x.lower() == variable_name][0]]
            else:
                error_message = "Variable not found"
                continue
            if len(variable) == 0:
                error_message = "Variable empty"
                continue

            try:
                if index is None:
                    df.loc[variable_key] = variable.reindex(years).values
                else:
                    variable.index = index_to_lowercase(variable.index)
                    index = tuple(x.lower() if isinstance(x, str) else x for x in index)
                    df.loc[variable_key] = variable.loc[index].reindex(years).values
            except KeyError:
                error_message = "Index not found"
        except Exception as e:
            error_message = f"{e}"
            print(f"Error getting {variable_key}: {error_message}")
        if error_message is not None:
            df.loc[variable_key] = error_message

    return df


import re

# Tidbits used to make regex-patterns more readable
open_bracket = r"[\(\[]"
close_bracket = r"[\)\]]"
brackets = r"[\(\)\[\]]"
no_brackets = r"[^\(\)\[\]]"
ident = r"[A-Za-z0-9_]{1,62}" # A valid symbol in GAMS is a letter followed by up to 62 letters, numbers or underscores. gamY also accepts starting with a number or underscore.
linebreak = r"(?:\r\n|\r|\n)"

# Top level patterns for all commands. Used to recursively parse script.
PATTERN_STRINGS = {
    # "env_variable": r"""(?:[^\s\*][^\n]*)?(\%(\S+?)\%)""",
    "env_variable": fr"""\%({ident})\%""",

    "user_function": fr"""
        @({ident})  # Function name
        \(
            ([^)]*?)  # Arguments
        \)
    """,

    "set": fr"""

                \$set(global|local|)
                \s+
                ({ident})
                \s+
                ([^\s\;]+)
            """,

    "eval": fr"""

                \$eval(global|local|)
                \s+
                ({ident})
                \s+
                ([^\;]+)
            """,

    "import": r"""
        \$Import
        \s+
        ([^\s;]+)                # File name
    """,

    "block": fr"\$Block\s+({ident})\s*({ident})?\s*(\$.+?)?\s*{linebreak}(.*?)\$EndBlock",

    "group": fr"\$Group(\+?)\s+({ident})\s+(.*?;)",  # 1) +? 2) name, 3) content

    "par_group": fr"\$P(?:ar)?Group(\+?)\s+({ident})\s+(.*?;)",  # 1) +? 2) name, 3) content

    "set_group": fr"\$S(?:et)?Group(\+?)\s+({ident})\s+(.*?;)",  # 1) +? 2) name, 3) content

    "display": r"""
        \$Display\s
            (.+?;)      # Variables and groups to be displayed
    """,

    "display_all": r"""
        \$Display_all
            (.+?;)      # Variables and groups to be displayed
    """,

    "model": fr"\$Model\s+({ident})\s+(.*?);",

    "solve": r"\$Solve\s+(.*?);",

    "fix": fr"""
                        (\$(?:UN)?FIX)          # Fix or unfix command
                        (?:{open_bracket} (
                            -? (?:\d+\.?\d*|INF|EPS) (?: \,\s*-?(?:\d+\.?\d*|INF|EPS) )?       # Optional arguments
                        ) {close_bracket})?
                        \s+
                        (.+?;)                  # Content
                    """,


    "if": r"""
                    \$If(?P<if_id>\d*)\s+
                    ([^:]*?)             # Condition
                    \:
                    (.*?)
                    \$EndIF(?P=if_id)\b
                """,

    "define_function": fr"""
                        \$Function(?P<function_id>\d*)\s+
                        ({ident})        # Name $1
                        {open_bracket}([^)\]]*){close_bracket}  # Arguments  $2
                        (?:\:)?\s+
                        (.*?)         # Expression $3
                        \$EndFunction(?P=function_id)\b
    """,

    "for_loop": r"""
                        \$For(?P<for_id>\d*)
                        \s+
                        (.+?) # The iterator
                        \s+
                        in
                        \s+
                        ([^\:]*?)  # The iterable
                        :
                        (.*?)
                        \$EndFor(?P=for_id)\b
    """,

    "loop": fr"""
                        \$Loop(?P<loop_id>\d*)
                        \s*
                        ({ident})               # Group name
                        [:]
                        (.*?)
                        \$EndLoop(?P=loop_id)\b
                    """,

    "replace": fr"""
                        \$Replace
                        {open_bracket}
                        ('.*?'|".*?")       # String to find
                        ,\s*
                        ('.*?'|".*?")       # Replacement string
                        (,\s*\d+)?          # Max replacements
                        {close_bracket}
                        (.*?)
                        \$EndReplace
                    """,

    "regex": fr"""
                        \$Regex(?P<regex_id>\d*)
                        {open_bracket}\s*
                        ('.*?'|".*?"|[^,]+)       # String to find
                        ,\s*
                        ('.*?'|".*?")             # Replacement string
                        (,\s*\d+)?                # Max replacements
                        {close_bracket}
                        (.*?)
                        \$EndRegex(?P=regex_id)\b
                    """,

    "eval_python": fr"""
        \$EvalPython
        (.*?) # Code
        \$EndEvalPython
    """
}

# Compile regex patterns
PATTERNS = {k: re.compile(v, re.VERBOSE | re.IGNORECASE | re.MULTILINE | re.DOTALL) for k, v in PATTERN_STRINGS.items()}
# Create combined pattern that matches any of the patterns
PATTERNS["Any"] = re.compile("|".join(PATTERN_STRINGS.values()), re.VERBOSE | re.IGNORECASE | re.MULTILINE | re.DOTALL)


PATTERNS["TopDown"] = re.compile("|".join((  # Remember to also add these to the list in gamY
        PATTERN_STRINGS["if"],
        PATTERN_STRINGS["for_loop"],
        PATTERN_STRINGS["loop"],
        PATTERN_STRINGS["define_function"],
    )),
    re.VERBOSE | re.IGNORECASE | re.MULTILINE | re.DOTALL
)



from pathlib import Path
import glob
import shutil

def find_root(root_file_name="paths.py"):
    """
    Find the root of the project by looking for a file named root_file_name.
    
    Parameters:
    root_file_name (str): The name of the file to look for (default is "paths.py").
    
    Returns:
    str: The path to the directory containing root_file_name, or None if not found.
    """
    current_path = Path.cwd()
    root_path = current_path.root

    while current_path != root_path:
        if (current_path / root_file_name).is_file():
            return str(current_path)
        current_path = current_path.parent

    return None

def clean_gams_temp_files():
    """
    Clean up temporary files and folders created by GAMS.
    """
    for dir in glob.glob("225*/"):
        shutil.rmtree(dir)


import os
import sys
import csv
import gamY
import re
import textwrap


def main():
  """
  Read model from gamY pickle
  Translate Equations to Gekko format
  Export equations to csv file
  """
  file_path = "../model/saved/model"  # sys.argv[1]
  file_dir, file_name = os.path.split(file_path)
  file_dir = os.path.abspath(file_dir)

  model_name = "M_base"  # sys.argv[2]

  db = gamY.Precompiler(file_path)
  db.read(file_name)

  with open(os.path.join(file_dir, file_name+'_equations.csv'), 'w', newline='') as file:
    writer = csv.writer(file, quoting=csv.QUOTE_ALL)
    writer.writerow(["Name", "Sets", "Conditionals", "Left hand side", "Right hand side", ""])
    for eq in db.blocks[model_name].values():
      LHS = textwrap.dedent(eq.LHS).replace("\n","")
      RHS = textwrap.dedent(eq.RHS).replace("\n","")
      while "  " in LHS:
        LHS = LHS.replace("  ", " ")
      while "  " in RHS:
        RHS = RHS.replace("  ", " ")
      writer.writerow([
        eq.name,
        eq.sets,
        eq.conditions,
        LHS,
        RHS,
      ])


t_only_pattern = re.compile(r"[(\[]t([+-]1)?[)\]]", re.IGNORECASE | re.MULTILINE | re.DOTALL)
t_end_pattern = re.compile(r",t([+-]1)?([)\]])", re.IGNORECASE | re.MULTILINE | re.DOTALL)
def remove_t(text):
  """Remove time dimension, t, as it is explicit in Gekko
  >>> remove_t("foo[a,t] =E= bar[t];")
  'foo[a] =E= bar;'
  """
  for m in t_only_pattern.finditer(text):
    if m.group(1):
      text = text.replace(m.group(0), f"[{m.group(1)}]")
    else:
      text = text.replace(m.group(0), "")
  for m in t_end_pattern.finditer(text):
    if m.group(1):
      text = text.replace(m.group(0), f")[{m.group(1)}]")
    else:
      text = text.replace(m.group(0), m.group(2))
  return text


sets_pattern = re.compile(r"[(\[](?:['\"a-z_,]|(?:\+1)|(?:-1))+[)\]]", re.IGNORECASE | re.MULTILINE | re.DOTALL)
def gekkofy_sets(text):
  """
  >>> gekkofy_sets("foo[a,t] =E= bar[t] + foobar[t+1];")
  'foo[#a] =E= bar + foobar[+1];'
  """
  text = remove_t(text)
  for match_text in sets_pattern.findall(text):
    sets = []
    for i in match_text[1:-1].split(","):
      if i[0] in "'\"":
        sets.append(i[1:-1])
      elif i[0] in "+-":
        sets.append(i)
      else:
        sets.append(f"#{i}")
    text = text.replace(match_text, "[{}]".format(','.join(sets)))
  return text


if __name__ == "__main__":
  import doctest
  main()



import os
from gams import *

input_path = os.path.abspath("..\\Model\\Saved\\previous_deep_calibration.gdx")
output_path = os.path.abspath("..\\Model\\Saved\\nye_navne.gdx")
replace_dict = {
  "agr": "lan",
  # "ene": "ene",
  "con": "byg",
  "ext": "udv",
  "hou": "bol",
  "man": "fre",
  "sea": "soe",
  "ser": "tje",
  "pub": "off",
}

ws = GamsWorkspace()
db = ws.add_database_from_gdx(input_path)
sets = [symbol for symbol in db if isinstance(symbol, GamsSet)]
variables = [symbol for symbol in db if isinstance(symbol, GamsVariable)]

for symbol in db:
  print(symbol.name)
  for r in symbol:
    new_keys = r.keys
    for i, k in enumerate(r.keys):
      for old_name, new_name in replace_dict.items():
        if k == old_name:
          new_keys[i] = new_name
    if new_keys != r.keys:
      if isinstance(symbol, GamsVariable):
        symbol.add_record(new_keys).level = r.level
      elif isinstance(symbol, GamsParameter):
        symbol.add_record(new_keys).value = r.value
      elif isinstance(symbol, GamsSet):
        symbol.add_record(tuple(new_keys))

db.export(output_path)


"""
Transform Excel work sheet into GAMS code to perform endo exo operations and change variable levels.
"""
import os
import sys
import pandas as pd
import xlwings as xw


def main():
  #  Read execution parameters
  workbook_path = os.path.abspath(sys.argv[1])
  assert os.path.splitext(workbook_path)[1].lower() in [".xls", ".xlsx", ".xlsm"], \
    f"{workbook_path} is not an Excel workbook"

  output_path = os.path.abspath(sys.argv[2])

  # Read Excel file
  print(f"Read endo_exo sheet from {workbook_path}")
  app = xw.App(visible=False)
  workbook = xw.Book(workbook_path)
  endo_exo = workbook.sheets["endo_exo"].range("B4:CK5000").options(pd.DataFrame, index=False).value
  app.quit()

  # Cast years in column header to ints
  header = list(endo_exo)
  year_columns_index = 4
  header[year_columns_index:] = [int(i) for i in header[year_columns_index:]]
  endo_exo.columns = header

  endo_exo_strings = endo_exo.apply(row_to_GAMS, axis=1, result_type="reduce", year_columns_index=year_columns_index)

  with open(output_path, "w") as file:
    print(f"Write to {output_path}")
    file.write("\n".join(endo_exo_strings))


def row_to_GAMS(row, year_columns_index, comment_char="#"):
  """Return GAMS code from row"""
  endo_exo_strings = []
  for interval in split_intervals(row[year_columns_index:]):
    if interval and not row['#']:
      interval_string = f"{interval[0]} <= t.val and t.val <= {interval[1]}"
      if row.endo:
        endo_exo_strings += [f"UNFIX {row.endo}$({interval_string});"]
      if row.exo:
        for t in interval:
          val = row.loc[t]
          name, elements = split_name(row.exo)
          var_input = row.exo.strip()
          if var_input.endswith("]")
            var = f"{var_input[:-1]},'{t}']"
            lagged = f"{var_input[:-1]},'{t-1}']"
          else
            var = f"{var_input}['{t}']"
            lagged = f"{var_input}['{t-1}']"
          if row.printcode == "dummy":
            pass
          elif row.printcode == "p":
            endo_exo_strings += [f"{var}$(t.val = {t}) = (1 + {val}/100) * ;"]
          elif row.printcode == "n":
            endo_exo_strings += [f"{var}$() = {val};"]
      else # Transfer comments
      endo_exo_strings += [f"{comment_char} {i}" for i in row.dropna()]

  return "\n".join(endo_exo_strings)

def split_name(input):
  """
  Split Gekko syntax name
  Return symbol name and a list of elements
  Example:
    >>> split_name("qI[#i,CON]")
    "qI", ["#i", "'CON'"]
  """
  if "[" in input:
    name, elements = input.strip().split("[", 1)
    elements = elements[:-1].split(",")
    elements = [e[1:] if (e[0] == "#") else f"'{e}'" for e in elements]
  else:
    name = input
    elements = []
  return name, elements


def split_intervals(series):
  interval = []
  for year in series.dropna().index:
    if (year - 1) not in interval:
      if interval:
        yield interval
        interval = []
    interval += [year]
  yield interval


def error(msg):
  exit(f"ERROR! {msg}")


if __name__ == "__main__":
  main()



import plotly.graph_objects as go
import math

dream_colors_rgb = {
  "DREAM": "rgb(245,82,82)",
  "REFORM": "rgb(66,180,224)",
  "Grøn REFORM": "rgb(92,210,114)",
  "SMILE": "rgb(255,155,75)",
  "MAKRO": "rgb(20,175,166)",
  "Plum": "rgb(188,173,221)",
  "Dark blue": "rgb(0,95,151)",
  "Maroon": "rgb(137,48,112)",
  "Dark gray": "rgb(70,70,76)",
  "Light gray": "rgb(230,230,232)",
}

dream_colors_hex = [
  "F55252", # DREAM
  "42B4E0", # REFORM
  "5CD272", # Grøn REFORM
  "FF9B4B", # SMILE
  "14AFA6", # MAKRO
  "BCADDD", # Plum
  "005F97", # Dark blue
  "893070", # Maroon
  "46464C", # Dark gray
  "E6E6E8", # Light gray
]

DPI = 96
DPCM = DPI / 2.54

large_figure_layout = dict(
  width = 15.5 * DPCM,
  # height = (plot_height + 3.8) * DPCM,
  margin_l = 1.5 * DPCM,
  margin_r = 0.5 * DPCM,
  margin_t = 1 * DPCM,
  # margin_b = 3.8 * DPCM,
)

small_figure_layout = dict(
  width = 7.6 * DPCM,
  # height = (plot_height + 5) * DPCM,
  margin_l = 1.5 * DPCM,
  margin_r = 0.5 * DPCM,
  margin_t = 1 * DPCM,
  # margin_b = 5 * DPCM,
)

def calculate_legend_height(trace_count, legend_item_height_cm=0.5, items_per_row=2):
  """
  Calculate the legend height based on the number of legend items,
  assuming a fixed number of items per row.
  
  :param trace_count: Number of traces (legend items) in the plot.
  :param legend_item_height_cm: Height per legend item in cm.
  :param items_per_row: Number of legend items per row.
  :return: Height of the legend in cm.
  """
  rows = math.ceil(trace_count / items_per_row)
  return rows * legend_item_height_cm

def calculate_figure_height(trace_count, base_height_cm=5, legend_item_height_cm=0.5, items_per_row=2):
  """
  Calculate the figure height based on the number of legend items.
  
  :param trace_count: Number of traces (legend items) in the plot.
  :param base_height_cm: Base height of the plot in cm.
  :param legend_item_height_cm: Height per legend item in cm.
  :param items_per_row: Number of legend items per row.
  :return: Total height of the figure in cm.
  """
  legend_height_cm = calculate_legend_height(trace_count, legend_item_height_cm, items_per_row)
  total_height_cm = base_height_cm + legend_height_cm
  return total_height_cm * DPCM

def create_dream_template(trace_count, items_per_row=2):
  """
  Create a Plotly template with dynamic height based on the number of traces.
  
  :param trace_count: Number of traces (legend items) in the plot.
  :param items_per_row: Number of legend items per row.
  :return: Plotly template.
  """
  figure_height = calculate_figure_height(trace_count, items_per_row=items_per_row)
  legend_height_cm = calculate_legend_height(trace_count, items_per_row=items_per_row)
  
  dream_layout = dict(
    colorway=list(dream_colors_rgb.values()),
    title_font_size=10,
    legend_orientation="h",
    legend_yanchor="top",
    legend_y=-legend_height_cm / DPCM,  # Position the legend below the plot
    legend_x=0.5,
    legend_xanchor="center",
    font=dict(size=10, family="Hind"),
    margin=dict(
      l=1.5 * DPCM,
      r=0.5 * DPCM,
      t=1 * DPCM,
      b=(legend_height_cm + 1) * DPCM  # Adjust bottom margin to accommodate the legend
    ),
    width=15.5 * DPCM,
    height=figure_height,
    xaxis=dict(
      title_font=dict(size=10),
      tickfont=dict(size=9),
      showgrid=True,
      ticks="outside",
      showline=True,
      mirror=True,
      zeroline=False,
    ),
    yaxis=dict(
      title_font=dict(size=10),
      tickfont=dict(size=9),
      showgrid=True,
      ticks="outside",
      showline=True,
      mirror=True,
      zeroline=False,
    ),
  )

  dream_template_data = dict(
    scatter=[
      dict(
        line=dict(width=1),
      )
    ]
  )

  dream_template = go.layout.Template(
    layout=dream_layout,
    data=dream_template_data,
  )
  
  return dream_template

# # Example usage
# trace_count = 10  # Number of traces (legend items)
# fig = go.Figure()

# # Add some sample traces
# for i in range(trace_count):
#   fig.add_trace(go.Scatter(
#     x=[1, 2, 3],
#     y=[i, i + 1, i + 2],
#     mode='lines+markers',
#     name=f'Trace {i}'
#   ))

# # Apply the custom template
# fig.update_layout(template=create_dream_template(trace_count))

# fig.show()